import torch

from astropy import constants, units


class Hydrodynamics:

    def __init__(self, gamma, mu, f_x=0.0, f_y=0.0, f_z=0.0, Lambda=0.0):

        # Set parameters
        self.gamma  = gamma
        self.mu     = mu
        self.f_x    = f_x
        self.f_y    = f_y
        self.f_z    = f_z
        self.Lambda = Lambda

        self.loss_keys = {
            'cont': 'hydro_cont',
            'momx': 'hydro_momx',
            'momy': 'hydro_momy',
            'momz': 'hydro_momz',
            'engy': 'hydro_engy',
        }



    def add_steady_state_loss(self, model, loss, log_rho, log_tmp, v_x, v_y, v_z):
        """
        Loss assuming steady state hydrodynamics, i.e. vanishing time derivatives in Euler's equations.
        """
        rho = torch.exp(log_rho)         
        tmp = torch.exp(log_tmp)
  
        # kB T / mu
        kBT_o_mu = (constants.k_B.si.value / self.mu) * tmp
    
        # log rho + log T
        log_rho_p_log_tmp = log_rho + log_tmp

        # Energy    
        eng = 0.5 * (v_x**2 + v_y**2 + v_z**2) + (self.gamma / (self.gamma - 1.0)) * kBT_o_mu

        # Continuity equation (steady state): div(ρ v) = 0
        loss_cont = model.diff_x(rho * v_x) + model.diff_y(rho * v_y) + model.diff_z(rho * v_z)

        # Momentum equation (steady state): v . grad(v) + grad(P) / rho = f
        loss_momx = v_x * model.diff_x(v_x) + v_y * model.diff_y(v_x) + v_z * model.diff_z(v_x) + kBT_o_mu * model.diff_x(log_rho_p_log_tmp) - self.f_x
        loss_momy = v_x * model.diff_x(v_y) + v_y * model.diff_y(v_y) + v_z * model.diff_z(v_y) + kBT_o_mu * model.diff_y(log_rho_p_log_tmp) - self.f_y
        loss_momz = v_x * model.diff_x(v_z) + v_y * model.diff_y(v_z) + v_z * model.diff_z(v_z) + kBT_o_mu * model.diff_z(log_rho_p_log_tmp) - self.f_z

        # Energy equation (steady state): div(u v) = 0
        loss_engy = rho * (model.diff_x(eng) * v_x + model.diff_y(eng) * v_y + model.diff_z(eng) * v_z) - self.Lambda

        # Compute the mean squares and add the losses
        loss[self.loss_keys['cont']] = torch.mean((loss_cont/     rho )**2)
        loss[self.loss_keys['momx']] = torch.mean((loss_momx/     v_x )**2)
        loss[self.loss_keys['momy']] = torch.mean((loss_momy/     v_y )**2)
        loss[self.loss_keys['momz']] = torch.mean((loss_momz/     v_z )**2)
        loss[self.loss_keys['engy']] = torch.mean((loss_engy/(rho*eng))**2)

        # Return losses
        return loss


import torch
import numpy             as np
import matplotlib.pyplot as plt

from astropy.io          import fits
from astropy             import units
from ipywidgets          import interact
from radio_beam          import Beam as RadioBeam
from astropy.convolution import convolve


# class Observation:    
#     def __init__(self):
#         print('Created an Observation object!')
        

class DataCube():
    """
    Data structure for a data cube of spectral line images.
    Helps to conveniently read and visualize spectral line images.
    """
    
    def __init__(self, fits_file):
        """
        Constructor for DataCube.

        Parameters
        ----------
        fits_file : str
            Path to the fits file containing the spectral line image.
        """
        self.read_fits_file(fits_file)
    
    
    def read_fits_file(self, fits_file):
        """
        Reader for fits files containing spectral line images.

        Parameters
        ----------
        fits_file : str
            Path to the fits file containing the spectral line image.
        """
        #Load the data
        with fits.open(fits_file) as hdu:
            # Extract data
            self.img = hdu[0].data[0].astype(np.float32)   # Store as 32 bit numbers to avoid too much memory
            # Extract header
            self.hdr = hdu[0].header                       

            # Extract the object name
            self.object = str(self.hdr['OBJECT']).strip()
            
            # Extract units for intensities
            self.I_unit = units.Unit(self.hdr['BUNIT'])
            
            # Extract units for axis
            self.x_unit = units.Unit(self.hdr['CUNIT1'])
            self.y_unit = units.Unit(self.hdr['CUNIT2'])
            self.f_unit = units.Unit(self.hdr['CUNIT3'])
            
            # Extract data from the header
            self.npix_x    =   int(self.hdr['NAXIS1'])
            self.npix_y    =   int(self.hdr['NAXIS2'])
            self.npix_f    =   int(self.hdr['NAXIS3'])

            self.pixsize_x = float(self.hdr['CDELT1']) * self.x_unit
            self.pixsize_y = float(self.hdr['CDELT2']) * self.y_unit
            self.pixsize_f = float(self.hdr['CDELT3']) * self.f_unit

            self.refpix_x  =   int(self.hdr['CRPIX1']) - 1
            self.refpix_y  =   int(self.hdr['CRPIX2']) - 1
            self.refpix_f  =   int(self.hdr['CRPIX3']) - 1

            self.refval_x  = float(self.hdr['CRVAL1']) * self.x_unit
            self.refval_y  = float(self.hdr['CRVAL2']) * self.y_unit
            self.refval_f  = float(self.hdr['CRVAL3']) * self.f_unit

            self.restfreq  = float(self.hdr['RESTFRQ'])

            # Verify image axis
            if (self.img.shape[1] != self.npix_x):
                raise ValueError('Number of pixels mismatch (AXIS 1, x).')
            if (self.img.shape[2] != self.npix_y):
                raise ValueError('Number of pixels mismatch (AXIS 2, y).')
            if (self.img.shape[0] != self.npix_f):
                raise ValueError('Number of pixels mismatch (AXIS 0, v).')
            if (self.hdr['CTYPE1'] != 'RA---SIN'):
                raise NotImplementedError("Don't know how to handle this.")
            if (self.hdr['CTYPE2'] != 'DEC--SIN'):
                raise NotImplementedError("Don't know how to handle this.")
            
            # Construct image axis
            self.xs = self.pixsize_x * (np.arange(self.npix_x) - self.refpix_x) + self.refval_x
            self.ys = self.pixsize_y * (np.arange(self.npix_y) - self.refpix_y) + self.refval_y
            self.fs = self.pixsize_f * (np.arange(self.npix_f) - self.refpix_f) + self.refval_f
            
            # Extract ranges for intensities
            self.img_min, self.img_max = np.min(self.img), np.max(self.img)
            
            # Extract ranges for axis
            self.x_min, self.x_max = np.min(self.xs), np.max(self.xs)
            self.y_min, self.y_max = np.min(self.ys), np.max(self.ys)
            self.f_min, self.f_max = np.min(self.fs), np.max(self.fs)
            

    def plot_channel_maps(self):
        """
        Plot channel maps for this observation.
        """
        # Define a plot function
        def plot(f):
            plt.figure(dpi=150)
            plt.imshow(self.img[f,:,:], vmin=self.img_min, vmax=self.img_max)
            
        # Return an interactive ipywidget
        return interact(plot, f=(0, self.img.shape[0]-1))
    
    
    def plot_xy_maps(self):
        """
        Plot x / y maps for this observation.
        (Alias for plot_channel_maps.)
        """
        return self.plot_channel_maps()
    
    
    def plot_xf_maps(self):
        """
        Plot x-axis / frequency maps for this observation.
        """
        # Define a plot function
        def plot(i):
            plt.figure(dpi=150)
            plt.imshow(self.img[:,i,:], vmin=self.img_min, vmax=self.img_max)
            
        # Return an interactive ipywidget
        return interact(plot, i=(0, self.img.shape[1]-1))
    
    
    def plot_yf_maps(self):
        """
        Plot y-axis / frequency maps for this observation.
        """
        # Define a plot function
        def plot(j):
            plt.figure(dpi=150)
            plt.imshow(self.img[:,:,j], vmin=self.img_min, vmax=self.img_max)
            
        # Return an interactive ipywidget
        return interact(plot, j=(0, self.img.shape[2]-1))
    
    
    def plot_spectrum(self):
        """
        Plot spectrum at a pixel for this observation.
        """
        # Define a plot function
        def plot(i,j):
            plt.figure(dpi=150)
            plt.step(self.fs, self.img[:,i,j])
            plt.ylim((self.img_min, self.img_max))
            plt.xlabel(f'frequency [{self.f_unit}]')
            plt.ylabel(f'intensity [{self.I_unit}]')

        # Return an interactive ipywidget
        return interact(plot,
                        i=(0, self.img.shape[1]-1),
                        j=(0, self.img.shape[2]-1) )


class Beam():
    
    def __init__(self, datacube):
        """
        Constructor for Beam.

        Parameters
        ----------
        datacube : DataCube
            DataCube object containing the spectral line image.
        """
        if abs(datacube.pixsize_x) != abs(datacube.pixsize_y):
            raise ValueError("Pixels are not square! Cannot handle non-square pixels!")

        # Create a radio-beam Beam object        
        self.object = RadioBeam.from_fits_header(datacube.hdr)

        # Extract the beam as a kernel
        self.kernel = self.object.as_kernel(datacube.pixsize_x)

        # Extract the beam as a torch kernel
        self.torch_kernel = self.get_torch_kernel(datacube.pixsize_x)


    def get_torch_kernel(self, pixsize):
        """
        Getter for the beam of the observation as a torch kernel.
        
        Parameters
        ----------
        pixsize : float
            Pixel size in the same units as the beam.

        Returns
        -------
        torch_kernel : torch.nn.Conv2d
            Torch kernel object.
        """
    
        # First create a kernel object
        kernel = self.object.as_kernel(pixsize) 

        torch_kernel = torch.nn.Conv2d(
            in_channels  = 1,
            out_channels = 1,
            kernel_size  = (kernel.array.shape[0], kernel.array.shape[1]),
            padding      = 'same',
            dtype        = torch.float64
        )

        torch_kernel.weight.data[0,0] = torch.from_numpy(kernel.array)
        torch_kernel.bias  .data      = torch.zeros_like(torch_kernel.bias)

        torch_kernel.weight.requires_grad_(False)
        torch_kernel.bias  .requires_grad_(False)

        return torch_kernel


    def apply(self, image):
        """
        Apply beam kernel to image.

        Parameters
        ----------
        image : np.ndarray
            Image to apply the beam to.

        Returns
        -------
        convolved_image : np.ndarray
            Input image convolved with the beam.
        """
        return convolve(image, self.kernel.array)
    
    
    def torch_apply(self, data):
        """
        Apply torch kernel.
        Assumes the last axis to contain the frequencies / channels.

        Parameters
        ----------
        data : torch.Tensor
            Data to apply the beam to.

        Returns
        -------
        dat : torch.Tensor
            Data convolved with the beam.
        """
        dat = torch.moveaxis(data, -1, 0).view(data.shape[2], 1, data.shape[0], data.shape[1])
        dat = self.torch_kernel(dat)
        dat = torch.moveaxis(dat.view(data.shape[2], data.shape[0], data.shape[1]), 0, -1)
        return dat


import torch
import matplotlib.pyplot as plt

from ipywidgets import interact


def plot_cube_2D(cube):
    """
    Plot a slice along the third axis through a 3D cube.

    Parameters
    ----------
    cube : torch.Tensor
        3D cube to plot.
    
    Returns
    -------
    out : None
    """
    vmin = cube.min().item()
    vmax = cube.max().item()
    def plot(z):
        plt.figure(dpi=150)
        plt.imshow(cube[:,:,z].T.data, vmin=vmin, vmax=vmax, origin='lower')
    return interact(plot, z=(0,cube.shape[2]-1))

    
def plot_spectrum(cube):
    """
    Plot spectrum at a pixel for this observation.

    Parameters
    ----------
    cube : torch.Tensor
        3D cube to plot the spectrum of.
    
    Returns
    -------
    out : None
    """
    # Define a plot function
    ymin = cube.min().item()
    ymax = cube.max().item()
    def plot(i,j):
        plt.figure(dpi=150)
        plt.plot(cube[i,j,:].data)
        plt.ylim((ymin, ymax))

    # Return an interactive ipywidget
    return interact(plot, i=(0, cube.shape[0]-1), j=(0, cube.shape[1]-1))


import torch
import numpy as np

from time             import time
from astroquery.lamda import Lamda, parse_lamda_datafile
from astropy          import constants, units
from pomme.utils   import get_molar_mass, print_var
from pomme.forward import image_along_last_axis as forward_image_along_last_axis


# Constants
CC  = constants.c  .si.value   # Speed of light       [m/s]
HH  = constants.h  .si.value   # Planck's constant    [J s]
KB  = constants.k_B.si.value   # Boltzmann's constant [J/K]
AMU = constants.u  .si.value   # Atomic mass unit     [kg]


class LineData():
    """
    Class structuring the line data.
    """
    
    def __init__(self, species_name=None, datafile=None, database='LAMDA'):
        """
        Constructor for linedata object.
        Either provide a datafile or a molecule name.
        
        Parameters
        ----------
        species_name : str
            Name of the lineproducing species.
        database : str
            Name of the type of line data base. (Default LAMDA.)
        """
        
        if database == 'LAMDA':
            # Use astroquery to read the LAMDA file
            if datafile is not None:
                collrates, radtransitions, enlevels = parse_lamda_datafile(datafile)
            elif species_name is not None:
                collrates, radtransitions, enlevels = Lamda.query(mol=species_name)
            else:
                raise ValueError("Either datafile or mol must be specified.")
        
            self.level  = np.array(enlevels['Level' ])
            self.energy = np.array(enlevels['Energy'])
            self.weight = np.array(enlevels['Weight'])
            self.J      = np.array(enlevels['J'     ])
            
            self.transition = np.array(radtransitions['Transition'])
            self.upper      = np.array(radtransitions['Upper'     ])
            self.lower      = np.array(radtransitions['Lower'     ])
            self.Einstein_A = np.array(radtransitions['EinsteinA' ])
            self.frequency  = np.array(radtransitions['Frequency' ])
            
            # Unit conversions to SI
            self.energy    *= 1.0E+2*HH*CC   # from [cm^-1] to [J]
            self.frequency *= 1.0e+9         # from [GHz]   to [Hz] 
            
            # Renumber levels and transitions from [1 - ] to [0 - ]
            self.level      -= 1
            self.transition -= 1
            self.upper      -= 1
            self.lower      -= 1
            
            # Derive Einstein B coefficients
            self.Einstein_Bs = self.Einstein_A * CC**2 / (2.0*HH*(self.frequency)**3)
            self.Einstein_Ba = self.weight[self.upper] / self.weight[self.lower] * self.Einstein_Bs
            
        else:
            raise NotImplementedError("Currently, only LAMDA data files are supported.")
            
            
class Line:
    """
    Spectral line class.
    """
    
    def __init__(self, species_name, transition, database='LAMDA', datafile=None, molar_mass=None):
        """
        Constructor for a line object.

        Parameters
        ----------
        species_name : str
            Name of the line producing species.
        transition : int
            Transition number of the line.
        database : str
            Name of the type of line data base. (Default LAMDA.)
        datafile : str
            Path to the datafile. (Default None.)
        molar_mass : float
            Molar mass of the species. (Default None.)
        """
        # Store the name of the species
        self.species_name = species_name
        # Determine the molar mass of the species
        if molar_mass is None:
            self.species_molar_mass = get_molar_mass(self.species_name)
        else:
            self.species_molar_mass = molar_mass
        # Store the transition number
        self.transition = transition
        # Extract the line data
        self.linedata = LineData(database=database, datafile=datafile, species_name=species_name.lower())
        
        self.upper       = self.linedata.upper       [self.transition]
        self.lower       = self.linedata.lower       [self.transition]
        self.frequency   = self.linedata.frequency   [self.transition]
        self.Einstein_A  = self.linedata.Einstein_A  [self.transition]
        self.Einstein_Bs = self.linedata.Einstein_Bs [self.transition]
        self.Einstein_Ba = self.linedata.Einstein_Ba [self.transition]
        
        self.J_upper = self.linedata.J[self.upper]
        self.J_lower = self.linedata.J[self.lower]
        
        self.energy = torch.from_numpy(self.linedata.energy)
        self.weight = torch.from_numpy(self.linedata.weight)
        
        self.description = f"{self.species_name}(J={self.J_upper}-{self.J_lower})"

        print(f"You have selected line:")
        print(f"    {self.description}")
        print(f"Please check the properties that were inferred:")
        print(f"    {'Frequency '       :<17} {self.frequency :0.9e}  Hz")
        print(f"    {'Einstein A coeff ':<17} {self.Einstein_A:0.9e}  1/s")
        print(f"    {'Molar mass'       :<17} {self.species_molar_mass:<15}  g/mol")

        
    def gaussian_width(self, temperature, v_turbulence):
        """
        Gaussian spectral line width.

        Parameters
        ----------
        temperature : torch.Tensor
            Temperature for which to evaluate the line width.
        v_turbulence : torch.Tensor
            Turbulent velocity for which to evaluate the line width.

        Returns
        -------
        out : torch.Tensor
            Tensor containing the Gaussian line width for the given temperature and turbulent velocity
        """
        # Compute convenience variables
        factor_1 = self.frequency / CC
        factor_2 = 2.0 * KB / (self.species_molar_mass * AMU)
        # Return the gaussian line width
        return factor_1 * torch.sqrt(factor_2*temperature + v_turbulence**2)


    def gaussian_profile(self, temperature, v_turbulence, freq):
        """
        Gaussian spectral line profile function.

        Parameters
        ----------
        temperature : torch.Tensor
            Temperature for which to evaluate the line profile.
        v_turbulence : torch.Tensor
            Turbulent velocity for which to evaluate the line profile.
        freq : torch.Tensor
            Frequency at which to evaluate the line profile.

        Returns
        -------
        out : torch.Tensor
            Tensor containing the Gaussian line profile for the given temperature and turbulent velocity
        """
        # Compute convenience variables
        inverse_width = 1.0 / self.gaussian_width(temperature, v_turbulence)
        factor        = 1.0 / np.sqrt(np.pi)
        # Mind the cellwise products
        shift  = torch.einsum("..., ...f -> ...f",          inverse_width, freq-self.frequency)
        result = torch.einsum("..., ...f -> ...f", factor * inverse_width, torch.exp(-shift**2))
        # Return the gaussian line profile
        return result
    
    
    def LTE_pops (self, temperature):
        """
        Relative LTE level populations for the given temperature.
    
        Parameters
        ----------
        temperature : torch.Tensor
            Temperature for which to evaluate the LTE level populations.
    
        Returns
        -------
        out : array_like
            Array containing the LTE level populations for the given temperature.
        """
        exponent = torch.einsum("i   ,  ... -> i...", -self.energy/KB, 1.0/temperature) 
        pop      = torch.einsum("i   , i... -> i...",  self.weight,    torch.exp(exponent))
        pop      = torch.einsum("i...,  ... -> i...",  pop,            1.0/torch.sum(pop, dim=0))
        # Return result
        return pop
    

    def emissivity_and_opacity_ij(self, pop):
        """
        Line opacity, not folded with the profile.
        
        Parameters
        ----------
        pop : torch.Tensor
            level populations
    
        Returns
        -------
        eta, chi : torch.Tensor, torch.Tensor
            Tensor containing the LTE emissivities and opacities for the given temperature.
        """
        # Compute the prefactor
        factor = HH * self.frequency / (4.0 * np.pi)
        
        # Compute the emissivity and opacity
        eta = factor *  self.Einstein_A  * pop[self.upper]
        chi = factor * (self.Einstein_Ba * pop[self.lower] - self.Einstein_Bs * pop[self.upper]) 
        
        # Return results
        return eta, chi


    def source_ij(self, pop):
        """
        Line source function

        Parameters
        ----------
        pop : torch.Tensor
            level populations
    
        Returns
        -------
        src : torch.Tensor
            Tensor containing the line source function (= emissivity / opacity).
        """
        return self.Einstein_A  * pop[self.upper] / (self.Einstein_Ba * pop[self.lower] - self.Einstein_Bs * pop[self.upper]) 

    

    def LTE_emissivity_and_opacity(self, abundance, temperature, v_turbulence, frequencies):
        """
        Line emissivity and opacity assuming LTE.
        
        Parameters
        ----------
        abundance : torch.Tensor
            Abundance distribution of the line producing species.
        temperature : torch.Tensor
            Temperature for which to evaluate the LTE level populations.
        v_turbulence : torch.Tensor
            Turbulent velocity for which to evaluate the line profile.
        frequencies : torch.Tensor
            Frequencies at which to evaluate the LTE emissivities and opacities.
    
        Returns
        -------
        eta, chi : torch.Tensor, torch.Tensor
            Tensor containing the LTE emissivities and opacities for the given temperature.
        """
        # Compute the prefactor
        factor = HH * self.frequency / (4.0 * np.pi)
        
        # t =- time()
        # Compute the LTE level populations
        pop = self.LTE_pops(temperature)
        # t += time()
        # print("LTE pop  ", t)
        
        
        # t =- time()
        # Compute the emissivity and opacity
        eta = factor *  self.Einstein_A  * pop[self.upper]
        chi = factor * (self.Einstein_Ba * pop[self.lower] - self.Einstein_Bs * pop[self.upper]) 
        # t += time()
        # print("Eins A B ", t)
        
        # t =- time()
        # Compute the (Gaussian) line profile
        profile = self.gaussian_profile(temperature, v_turbulence, frequencies)
        # t += time()
        # print("profile  ", t)
        
        # t =- time()
        # Fold the emessivities and opacities with the profile and the number abundance
        eta = torch.einsum("..., ...f -> ...f", eta*abundance, profile)
        chi = torch.einsum("..., ...f -> ...f", chi*abundance, profile)
        # t += time()
        # print("multiply ", t)
        
        # Return results
        return (eta, chi)
    

    # def optical_depth_along_last_axis_slow(self, chi_ij, abundance, temperature, v_turbulence, velocity_los, frequencies, dx):
    #     """
    #     Line optical depth along the last axis.
    #     """
    #     sqrt_pi = np.sqrt(np.pi)

    #     # Compute inverse line width
    #     inverse_width = 1.0 / self.gaussian_width(temperature=temperature, v_turbulence=v_turbulence)
  
    #     # Get the index of the last spatial axis
    #     last_axis = abundance.dim() - 1

    #     # Compute the Doppler shift for each element
    #     shift = 1.0 + velocity_los * (1.0 / CC)

    #     # Create freqency tensor in the rest frame for each element
    #     freqs_restframe = torch.einsum("..., f -> ...f", shift, frequencies)

    #     # Define the a and b (tabulated) functions
    #     a = (1.0/sqrt_pi) * inverse_width * chi_ij * abundance
    #     a = torch.einsum("...,    f -> ...f", a, torch.ones_like(frequencies))
    #     b = torch.einsum("..., ...f -> ...f", inverse_width, freqs_restframe - self.frequency)

    #     a0 = a[..., :-1, :]
    #     a1 = a[..., 1: , :]
    
    #     b0 = b[..., :-1, :]
    #     b1 = b[..., 1: , :]

    #     b10 = b1 - b0

    #     # threashhold differentiating the two regimes (large and small Doppler shift)
    #     shift_threshold = 1.0e-3
    
    #     # Define the masks for the threashold    
    #     A = torch.Tensor(torch.abs(b10) >  shift_threshold)
    #     B = torch.Tensor(torch.abs(b10) <= shift_threshold)

    #     dtau = torch.empty_like(b10)
        
    #     dtau[A]  =           (      a1[A] -       a0[A]) * (torch.exp(-b0[A]**2) - torch.exp(-b1[A]**2))
    #     dtau[A] += sqrt_pi * (b0[A]*a1[A] - b1[A]*a0[A]) * (torch.erf( b0[A]   ) - torch.erf( b1[A]   ))
    #     dtau[A] *= 0.5 / b10[A]**2

    #     dtau[B]  = (1.0/ 2.0) * (a0[B] +     a1[B])
    #     dtau[B] -= (1.0/ 3.0) * (a0[B] + 2.0*a1[B]) * b0[B]                        * b10[B]   
    #     dtau[B] += (1.0/12.0) * (a0[B] + 3.0*a1[B]) *         (2.0*b0[B]**2 - 1.0) * b10[B]**2
    #     dtau[B] -= (1.0/30.0) * (a0[B] + 4.0*a1[B]) * b0[B] * (2.0*b0[B]**2 - 3.0) * b10[B]**3
    #     dtau[B] *= torch.exp(-b0[B]**2)
        
    #     dtau *= dx

    #     tau = torch.empty_like(b)
    #     tau[...,  0 , :] = 0.0
    #     tau[..., +1:, :] = torch.cumsum(dtau, dim=last_axis)

    #     return dtau, tau


    def optical_depth_along_last_axis(self, chi_ij, abundance, temperature, v_turbulence, velocity_los, frequencies, dx):
        """
        Line optical depth along the last axis.

        Parameters
        ----------
        chi_ij : torch.Tensor
            Line opacity distribution.
        abundance : torch.Tensor
            Abundance distribution of the line producing species.
        temperature : torch.Tensor
            Temperature distribution of the line producing species.
        v_turbulence : torch.Tensor
            Turbulent velocity distribution.
        velocity_los : torch.Tensor
            Line of sight velocity distribution.
        frequencies : torch.Tensor
            Frequencies at which to compute the optical depth.
        dx : torch.Tensor
            Grid spacing along the line of sight.

        Returns
        -------
        dtau : torch.Tensor
            Tensor containing the differential optical depth.
        tau : torch.Tensor
            Tensor containing the cumulative optical depth.
        """
        sqrt_pi = np.sqrt(np.pi)

        # Compute inverse line width
        inverse_width = 1.0 / self.gaussian_width(temperature=temperature, v_turbulence=v_turbulence)
  
        # Get the index of the last spatial axis
        last_axis = abundance.dim() - 1

        # Compute the Doppler shift for each element
        shift = 1.0 + velocity_los * (1.0 / CC)

        # Create freqency tensor in the rest frame for each element
        freqs_restframe = torch.einsum("..., f -> ...f", shift, frequencies)

        # Define the a and b (tabulated) functions (Note: we absorb dx in the definition of a for efficiency)
        a = (1.0/sqrt_pi) * inverse_width * chi_ij * abundance
        b = torch.einsum("..., ...f -> ...f", inverse_width, freqs_restframe - self.frequency)

        # if a.isnan().any():
        #     raise Warning("NaNs in a.")
        # if b.isnan().any():
        #     raise Warning("NaNs in b.")

        expb = torch.exp(-b**2) 
        erfb = torch.erf( b   ) 

        a0 = a[..., :-1] * dx
        a1 = a[..., 1: ] * dx
    
        b0 = b[..., :-1, :]
        b1 = b[..., 1: , :]

        expb0 = expb[..., :-1, :]
        expb1 = expb[..., 1: , :]
        
        erfb0 = erfb[..., :-1, :]
        erfb1 = erfb[..., 1: , :]

        b10 = b1 - b0
        a01 = a0 + a1

        sp_a1b0 = torch.einsum("..., ...f -> ...f", sqrt_pi * a1, b0)
        sp_a0b1 = torch.einsum("..., ...f -> ...f", sqrt_pi * a0, b1)

        # if a1.isnan().any():
        #     raise Warning("NaNs in a1.")
        # if a0.isnan().any():
        #     raise Warning("NaNs in a0.")
        # if b0.isnan().any():
        #     raise Warning("NaNs in b0.")
        # if b1.isnan().any():
        #     raise Warning("NaNs in b1.")
        # if sp_a1b0.isnan().any():
        #     raise Warning("NaNs in sp_a1b0.")
        # if sp_a0b1.isnan().any():
        #     raise Warning("NaNs in sp_a0b1.")

        dtau  = torch.einsum("..., ...f -> ...f", a1 - a0, expb0 - expb1)
        dtau += (sp_a1b0 - sp_a0b1) * (erfb0 - erfb1)
        dtau *= (0.5 / (b10**2 + 1.0e-30))
        # note that the regularizer 1.0e-30 is never going to be significant
        # however it is essential to avoid nans in backprop (see https://github.com/Magritte-code/pomme/issues/2)

        # if dtau.isnan().any():
        #     raise Warning("NaNs in dtau before mask.")

        # Create a mask for the region where the above calculation might have numerical issues
        mask_threshold = 1.0e-4
        mask           = torch.Tensor(torch.abs(b10) < mask_threshold)

        # Use a (second order) Taylor expansion in b10 for the masked region
        dtau[mask] = (   torch.einsum("..., ...f", (1.0/2.0) *  a01      , expb0           ) \
                       - torch.einsum("..., ...f", (1.0/3.0) * (a01 + a1), expb0 * b0 * b10) )[mask]

        # if dtau.isnan().any():
        #     raise Warning("NaNs in dtau after mask.")

        tau = torch.empty_like(b)
        tau[...,  0 , :] = 0.0
        tau[..., +1:, :] = torch.cumsum(dtau, dim=last_axis)

        return dtau, tau

    
    def image_along_last_axis(self, pop, abundance, temperature, v_turbulence, velocity_los, frequencies, dx, img_bdy):
        """
        Create an image along the last data axis.

        Parameters
        ----------
        pop : torch.Tensor
            Level populations for the line producing species.
        abundance : torch.Tensor
            Abundance distribution of the line producing species.
        temperature : torch.Tensor
            Temperature distribution of the line producing species.
        v_turbulence : torch.Tensor
            Turbulent velocity distribution.
        velocity_los : torch.Tensor
            Line of sight velocity distribution.
        frequencies : torch.Tensor
            Frequencies at which to image the model.
        dx : torch.Tensor
            Grid spacing along the line of sight.
        img_bdy : torch.Tensor
            Boundary conditions for the image.
    
        Returns
        -------
        img : torch.Tensor
            Tensor containing the image of the model.
        """
        # Compute the line emissivity and opacity from the level populations
        eta_ij, chi_ij = self.emissivity_and_opacity_ij(pop=pop)

        # if eta_ij.isnan().any() or chi_ij.isnan().any():
        #     raise Warning("NaNs in emissivity or opacity.")

        dtau, tau = self.optical_depth_along_last_axis(
            chi_ij       = chi_ij,
            abundance    = abundance,
            temperature  = temperature,
            v_turbulence = v_turbulence,
            velocity_los = velocity_los,
            frequencies  = frequencies,
            dx           = dx
        )

        # if dtau.isnan().any() or tau.isnan().any():
        #     raise Warning("NaNs in dtau or tau.")

        src = eta_ij / chi_ij
        img = forward_image_along_last_axis(src=src, dtau=dtau, tau=tau, img_bdy=img_bdy)

        # if img.isnan().any():
        #     raise Warning("NaNs in image.")

        return img

    # @torch.compile
    def LTE_image_along_last_axis(self, abundance, temperature, v_turbulence, velocity_los, frequencies, dx, img_bdy):
        """
        Create an image along the last data axis, assuming LTE level populations.

        Parameters
        ----------
        abundance : torch.Tensor
            Abundance distribution of the line producing species.
        temperature : torch.Tensor
            Temperature distribution of the line producing species.
        v_turbulence : torch.Tensor
            Turbulent velocity distribution.
        velocity_los : torch.Tensor
            Line of sight velocity distribution.
        frequencies : torch.Tensor
            Frequencies at which to image the model.
        dx : torch.Tensor
            Grid spacing along the line of sight.
        img_bdy : torch.Tensor
            Boundary conditions for the image.
        
        Returns
        -------
        img : torch.Tensor
            Tensor containing the image of the model.
        """

        pop = self.LTE_pops(
            temperature = temperature
        )

        # print_var('pop         ', pop)
        # print_var('abundance   ', abundance)
        # print_var('temperature ', temperature)
        # print_var('v_turbulence', v_turbulence)
        # print_var('velocity_los', velocity_los)
        # print_var('frequencies ', frequencies)

        img = self.image_along_last_axis(
            pop          = pop,
            abundance    = abundance,
            temperature  = temperature,
            v_turbulence = v_turbulence,
            velocity_los = velocity_los,
            frequencies  = frequencies,
            dx           = dx,
            img_bdy      = img_bdy
        )

        # print_var('img', img)

        return img


    def freq_to_velo(self, freq, unit='m/s'):
        """
        Convert frequencies with respect to this line to velocities.

        Parameters
        ----------
        freq : torch.Tensor
            Frequency tensor to convert.
        unit : str
            Unit of the velocity. (Default 'm/s'.)

        Returns
        -------
        out : torch.Tensor
            Tensor containing the velocities corresponding to the given frequencies.
        """
        return (freq / (self.frequency * units.Hz) - 1.0) * constants.c.to(unit).value


import torch


def image_along_last_axis(src, dtau, tau, img_bdy):
    """
    Create syntehtic image along the last data axis.

    Parameters
    ----------
    src : torch.Tensor
        Source function for the radiation transport equation.
    dtau : torch.Tensor
        Optical depth difference between two consequtive points.
    tau : torch.Tensor
        Optical depth.
    img_bdy : torch.Tensor
        Boundary condition for the image.

    Returns
    -------
    img : torch.Tensor
        Synthetic image along the last axis.
    """
    
    # Check dimensionality of the input
    assert src.dim() == tau.dim()-1, "Tensor src should only have spatial dimensions, no frequency!"

    # Get the index of the last spatial axis
    last_axis = src.dim() - 1

    src_0 = src[..., :-1]
    src_1 = src[..., +1:]

    exp_minus_tau = torch.exp(-tau)

    emt_0 = exp_minus_tau[..., :-1, :]
    emt_1 = exp_minus_tau[..., +1:, :]

    # threshold differentiating the two optical dpeth regimes
    mask_threshold = 1.0e-4
    mask = torch.Tensor(dtau < mask_threshold)

    # Case a: dtau > threshold 
    result  = torch.einsum("..., ...f -> ...f", src_0, emt_1 - emt_0 * (1.0 - dtau))
    result += torch.einsum("..., ...f -> ...f", src_1, emt_0 - emt_1 * (1.0 + dtau))
    result /= (dtau + 1.0e-30)
    # note that the regularizer 1.0e-30 is never going to be significant
    # however it is essential to avoid nans in backprop (see https://github.com/Magritte-code/pomme/issues/2)
    
    # Use a Taylor expansion for small dtau
    cc     = (1.0/2.0) * dtau
    fac_0  = cc.clone() 
    fac_1  = cc.clone()
    cc    *= (1.0/3.0) * dtau 
    fac_0 += cc
    fac_1 -= cc
    cc    *= (1.0/4.0) * dtau 
    fac_0 += cc
    fac_1 -= cc

    result[mask] = (  torch.einsum("..., ...f -> ...f", src_0, emt_0 * fac_0) \
                    + torch.einsum("..., ...f -> ...f", src_1, emt_1 * fac_1) )[mask]

    img = img_bdy * exp_minus_tau[..., -1, :] + torch.sum(result, dim=last_axis)

    return img


from pomme.utils import get_molar_mass


def test_get_molar_mass():
    """
    Test the molar mass function.
    """
    assert          get_molar_mass('H') == 1.00794
    assert 18.015 < get_molar_mass('H2O')                    < 18.016
    assert 98.07  < get_molar_mass('H2SO4')                  < 98.08
    assert 386    < get_molar_mass('CF3OCF(CF3)CF2OCF2OCF3') < 386.1
    assert 159.6  < get_molar_mass('Fe2O3')                  < 159.7


import torch
import torch.nn          as nn
import numpy             as np
import matplotlib.pyplot as plt


class Loss():
    """
    A convenience class to store losses.
    """
    
    def __init__(self, keys=[]):
        """
        Constructor for loss.

        Parameters
        ----------
        keys : list
            List of keys for the losses.
        """
        # Store keys
        self.keys = keys
        # Provide entries for all losses, norm, and weight
        self.loss   = {key:torch.zeros(1) for key in self.keys}
        self.norm   = {key:1.0            for key in self.keys} 
        self.weight = {key:1.0            for key in self.keys} 
        self.losses = {key:[]             for key in self.keys}
        # Provide entries for tot loss
        self.loss_tot   = torch.zeros(1)
        self.losses_tot = []


    def reset(self):
        """
        Reset all losses and remove all stored losses.
        """
        # Provide entries for all losses, norm, and weight
        self.losses = {key:[] for key in self.keys}
        # Provide entries for tot loss
        self.loss_tot   = torch.zeros(1)
        self.losses_tot = []

    
    def __getitem__(self, key):
        """
        Getter for variables (vars). Allows the use of [] operators.

        Parameters
        ----------
        key : str
            Key of the variable to be returned.

        Returns
        -------
        torch.Tensor
            The requested variable.
        """
        return self.loss[key]
    
    
    def __setitem__(self, key, value):
        """
        Setter for variables (vars). Allows the use of [] operators.

        Parameters
        ----------
        key : str
            Key of the variable to be set.
        value : torch.Tensor
            Value to be set.

        Raises
        ------
        ValueError
            If the value is not a torch.Tensor.
        """
        if isinstance(value, torch.Tensor):
            # weight
            value *= self.weight[key]
            # normalise
            value *= self.norm[key]
            # store
            self.loss[key] = value
            self.losses[key].append(value.item())
        else:
            raise ValueError("Only torch.Tensor is supported.")


    def renormalise(self, key):
        """
        Reset the norm to one over the current loss.

        Parameters
        ----------
        key : str
            Key of the variable to be renormalised.
        """
        self.norm[key] *= 1.0 / self.loss[key].item()


    def renormalise_all(self):
        """
        Renormalise all losses.
        """
        for key in self.keys:
            self.renormalise(key)

            
    def tot(self):
        """
        Return the total loss.
        """
        # Initialise the total loss
        self.loss_tot = torch.zeros(1)
        # Add all losses
        for l in self.loss.values():
            self.loss_tot += l
        # Add tot list
        self.losses_tot.append(self.loss_tot.item())
        # Return the total loss
        return self.loss_tot
    
    
    def plot(self):
        """
        Plot the evolution of the losses.
        """
        plt.figure(dpi=130)
        # Plot the total loss
        plt.plot(self.losses_tot, label='tot', linestyle=':', c='k')
        # Plot all constituent losses
        for key in self.losses.keys():
            plt.plot(self.losses[key], label=key)
        plt.yscale('log')
        plt.legend()
        
        
def haar_loss_1D(arr):
    """
    Loss based on the Haar wavelet transform.
    """
    loss = torch.zeros(1)
    while len(arr) > 1:
        loss += nn.functional.mse_loss(arr[0::2], arr[1::2]) * len(arr)
        arr = 0.5*(arr[0::2] + arr[1::2])
    return loss


def fourier_loss_1D(arr):
    """
    Loss based on the (1D) Fourier transform.
    """
    # Take the (real) Fast Fourier Transform
    fft = torch.abs(torch.fft.rfft(arr))
    # Store the size of the resulting transform
    size = fft.size(0)
    # Define the weights
    wts = torch.arange(size)
    # Compute the mean square loss
    loss = torch.mean((fft*wts)**2)
    # Return the loss
    return loss


def diff_loss(arr):
    """
    Differential loss, quantifying the local change in a variable along the cartesian axes.
    """
    loss = torch.zeros(1)
    for d in range(arr.dim()):
        loss += torch.mean(torch.diff(arr, dim=d)**2)
    return loss


class SphericalLoss:
    """
    Copmutes the deviation from spherical symmetry.
    This is quantified as the variacne of the data in each radial bin.
    """
    def __init__(self, model, origin='centre', weights=None):
        """
        Constructor for the spherical loss.

        Parameters
        ----------
        model : TensorModel
            TensorModel object on which the loss should be applied.
        origin : array_like or str
            Origin of the spherical coordinates, given as indices of the origin
            of the coordinate system (can be float).
            The dimension of the origin should match the dimension of the model.
            If 'centre', the origin is set to the centre of the model.
        weights : torch.Tensor, optional
            Weights for the radial bins. If None, the weights are set to one.
        """
        # Get radial coordinates w.r.t. the given origin.
        r = model.get_radius(origin=origin)
        # Define the number of radial bins.
        self.N = int(np.mean(model.shape)//2)
        # Define the boundaries of the radial bins.
        r_min = r.min()
        r_max = r.max()
        # Extract the indices of the radial bins.
        self.r_ind = ((r - r_min) *  ((self.N - 1) / (r_max - r_min))).astype(np.int64)
        # Extract the spherical masks (which mask everything but radial bin 'i').
        self.masks = [self.r_ind==i for i in range(self.N)]
        # Set the weights, if necessary.
        if weights is None:
            self.weights = torch.ones(self.N)
        else:
            self.weights = weights


    def eval(self, var):
        """
        Evaluate the spherical loss.

        Parameters
        ----------
        var : torch.Tensor
            Variable for which the loss should be evaluated.

        Returns
        -------
        torch.Tensor
            The spherical loss for the given variable.
        """
        # Compute the standard deviation of the variable data in each radial bin.
        sph_std = torch.zeros(self.N)
        for i in range(self.N):
            sph_std[i] = var[self.masks[i]].std()
        return torch.mean(self.weights * sph_std)


# def steady_state_continuity_loss(model, rho, v_x, v_y, v_z):
#     """
#     Loss assuming steady state hydrodynamics, i.e. vanishing time derivatives in Euler's equations.
#     """
#     # Continuity equation (steady state): div(ρ v) = 0
#     loss_cont = model.diff_x(rho * v_x) + model.diff_y(rho * v_y) + model.diff_z(rho * v_z)
#     # Squared average over the model
#     loss_cont = torch.mean((loss_cont / rho)**2)
#     return loss_cont


import torch

from pomme.model import TensorModel


def test_constructor_1D():
    model = TensorModel(
        keys  = ['CO'],
        sizes = (1.0,),
        shape = (100,),
    )


def test_constructor_2D():
    model = TensorModel(
        keys  = ['CO'],
        sizes = (1.0, 1.0),
        shape = (100, 100),
    )


def test_constructor_3D():
    model = TensorModel(
        keys  = ['CO'],
        sizes = (1.0, 1.0, 1.0),
        shape = (100, 100, 100),
    )
    

def test_is_field_1D():
    # Create a 3D model
    model = TensorModel(
        keys  = ['CO'],
        sizes = (1.0),
        shape = (100),
    )
    # Create a non-field variable
    model['non_field_var'] = torch.Tensor([1.23])
    # Check is_field
    assert model.is_field('CO')            == True
    assert model.is_field('non_field_var') == False
    

def test_is_field_3D():
    # Create a 3D model
    model = TensorModel(
        keys  = ['CO'],
        sizes = (1.0, 1.0, 1.0),
        shape = (100, 100, 100),
    )
    # Create a non-field variable
    model['non_field_var'] = torch.Tensor([1.23])
    # Check is_field
    assert model.is_field('CO')            == True
    assert model.is_field('non_field_var') == False


from astropy             import units
from astropy.coordinates import SkyCoord
from astroquery.simbad   import Simbad


class AstroObject():
    """
    Class to store data about an astrophysical object,
    such as distance, radial velocity, and sky coordinates.
    """

    @units.quantity_input(distance='length', radial_velocity='velocity')
    def __init__(self, name, distance=None, radial_velocity=None):
        """
        Constructor for AstroObject.

        Parameters
        ----------
        name : str
            Name of the astrophysical object.
        distance : astropy.units.quantity.Quantity object, optional
            Distance to the astrophysical object.
        radial_velocity : astropy.units.quantity.Quantity object, optional
            Radial velocity of the astrophysical object.

        Raises
        ------
        RuntimeError
            If the object name is not found in the SIMBAD database.
        """
        # Set name
        self.name = name
        
        # Get coordinates in the sky
        self.coordinates = SkyCoord.from_name(self.name)
        
        # Query SIMBAD for data
        simbadGetter = SimbadGetter(self.name)

        # Set distance
        if distance is not None:
            self.distance = distance
        else:
            self.distance = simbadGetter.get_distance()
            
        # Set radial velocity
        if radial_velocity is not None:
            self.radial_velocity = radial_velocity
        else:
            self.radial_velocity = simbadGetter.get_radial_velocity()

            
class SimbadGetter():
    """
    Class to query data from the SIMBAD data base.
    """
    def __init__(self, name: str):
        """
        Constructor for Simbad getter class.

        Parameters
        ----------
        name : str
            Name of the astrophysical object.

        Raises
        ------
        RuntimeError
            If the object name is not found in the SIMBAD database.
        """
        # Store the object name.
        self.name = name
        
        # Add votable fields to get the relevant data
        Simbad.add_votable_fields('plx', 'distance', 'rv_value')
        
        # Query Simbad based on the name
        self.sim = Simbad.query_object(self.name)
        
        # Check if the query was succesful
        if not self.sim:
            raise RuntimeError("No object with name", self.name, "was found in Simbad.")

            
    def get_distance(self):
        """
        Getter for the distance to the astrophysical object, using Simbad.
        Code adapted from:
        https://gist.github.com/elnjensen/ce2367faf0d876def1ff68b6154e102b

        Returns
        -------
        dist : astropy.units.quantity.Quantity object
            Distance to astrophysical object, from Simbad.
        """
        # Extract parallax
        plx = self.sim['PLX_VALUE']
        # Extract other distance
        dst = self.sim['Distance_distance']
    
        # Prefer the parallax distance if available,
        if not plx.mask[0]:
            print("Using the parallax distance.")
            # Convert parallax to distance
            return plx.to(units.pc, equivalencies=units.parallax())[0]
                      
        # but if no parallax we'll take any other distance.
        elif not dst.mask[0]:
            print("Using the distance from ", self.sim['Distance_bibcode'][0])
            # Extract distance and unit
            return (dst[0] * units.Unit(self.sim['Distance_unit'][0])).to(units.pc)

        # If there is no parallax or distance, we throw an error...
        else: 
            raise RuntimeError("No distance available in Simbad.")
        
        
    def get_radial_velocity(self):
        """
        Getter for the radial velocity of the astrophysical object, using Simbad.

        Returns
        -------
        velo : astropy.units.quantity.Quantity object
             Radial velocity of the astrophysical object, from Simbad.

        Raises
        ------
        RuntimeError
            If the object has no radial velocity in the SIMBAD database.
        """
        # Extract radial velocity
        vel = self.sim['RV_VALUE']
    
        # Check if a value was returned
        if not vel.mask[0]:
            # Convert to SI units
            return vel.to(units.m/units.s)[0]

        # If there is no radial velocity, we throw an error...
        else: 
            raise RuntimeError("No radial velocity available in Simbad.")


import torch
import numpy as np

from os                import path
from pomme.observation import DataCube, Beam

# Get the path to this folder
this_folder = path.dirname(path.abspath(__file__))


def test_torch_vs_numpy_beam_kernel():
    # Read the data cube
    data = DataCube(f'{this_folder}/data/header_only.fits')
    # Extract the beam
    beam = Beam(data)

    N = 50

    # Create a random image
    img_torch = torch.rand(N,N,1, dtype=torch.float64)
    img_numpy = img_torch[:,:,0].numpy()

    # Convolve once using the astropy functionality and once with the torch implementation
    img_torch_conv = beam.torch_apply(img_torch)[:,:,0].numpy()
    img_numpy_conv = beam.apply      (img_numpy)

    # Compute the absolute relative difference
    err = np.abs(2.0 * (img_torch_conv - img_numpy_conv) / (img_torch_conv + img_numpy_conv))

    assert err.max() < 5.0e-14


import torch
import numpy             as np
import matplotlib.pyplot as plt

from astropy    import units, constants
from torch.nn   import functional
from ipywidgets import interact

# Cosmic Microwave Background temperature
T_CMB = 2.725


def planck(temperature, frequency):
    '''
    Planck function for thermal radiation.

    Parameters
    ----------
    temperature : float
        Temperature at which to evaluate the intensity.
    frequency : float
        Frequency at which to evaluate the intensity.

    Returns
    -------
    out : float
        Planck function evaluated at the frequency for the given temperature.
    '''
    # Specify constants
    h  = constants.h  .si.value
    c  = constants.c  .si.value
    kb = constants.k_B.si.value
    # Return planck function
    return 2.0*h/c**2 * np.power(frequency, 3) / np.expm1(h*frequency/(kb*temperature))


def print_var(name, var):
    """
    Print the min, mean, and max of a PyTorch tensor variable.

    Parameters
    ----------
    name : str
        Name of the variable.
    var : torch.Tensor
        Variable to print the min, mean, and max of.

    Returns
    -------
    out : None
    """
    print(f"{name} {var.min().item():+1.2e} {var.mean().item():+1.2e} {var.max().item():+1.2e}")


def interpolate(inp, size, mode='nearest'):
    """
    Interpolate a tensor to a new size.

    Parameters
    ----------
    inp : torch.Tensor
        Input tensor to interpolate.
    size : tuple
        New size of the tensor.
    mode : str
        Interpolation mode.
    
    Returns
    -------
    out : torch.Tensor
        Interpolated tensor.
    """
    # Reshape to add batch ids and channels
    # (Because that is what torch.nn.functional.interpolate expects.)
    res = inp.view((1, 1) + inp.size())
    # Interpolate
    res = functional.interpolate(res, size=size, mode=mode)
    # Return without batch ids and channels
    return res.view(size)


@units.quantity_input(angle='angle', distance='length')
def convert_angular_to_spatial(angle, distance):
    """
    Convert angles to distances assuming a certain distance.

    Parameters
    ----------
    angle : astropy.units.Quantity
        Angle to convert.
    distance : astropy.units.Quantity
        Distance to assume.

    Returns
    -------
    out : astropy.units.Quantity
        Distance corresponding to the angle at the given distance.
    """
    angle    = angle   .to(units.arcsec).value
    distance = distance.to(units.pc    ).value
    return angle * distance * units.au


# https://gist.github.com/elibroftw/22e3b4c1eb7fa0a6c83d099d24200f95
# =================================
# Molar Mass Calculator
# Author: Elijah Lopez
# Version: 1.2.1
# Last Updated: April 4th 2020
# Created: July 8th 2017
# Python Version: 3.6+
# =================================
MM_of_Elements = {
    'H': 1.00794, 'He': 4.002602, 'Li': 6.941, 'Be': 9.012182, 'B': 10.811, 'C': 12.0107, 'N': 14.0067,
    'O': 15.9994, 'F': 18.9984032, 'Ne': 20.1797, 'Na': 22.98976928, 'Mg': 24.305, 'Al': 26.9815386,
    'Si': 28.0855, 'P': 30.973762, 'S': 32.065, 'Cl': 35.453, 'Ar': 39.948, 'K': 39.0983, 'Ca': 40.078,
    'Sc': 44.955912, 'Ti': 47.867, 'V': 50.9415, 'Cr': 51.9961, 'Mn': 54.938045,
    'Fe': 55.845, 'Co': 58.933195, 'Ni': 58.6934, 'Cu': 63.546, 'Zn': 65.409, 'Ga': 69.723, 'Ge': 72.64,
    'As': 74.9216, 'Se': 78.96, 'Br': 79.904, 'Kr': 83.798, 'Rb': 85.4678, 'Sr': 87.62, 'Y': 88.90585,
    'Zr': 91.224, 'Nb': 92.90638, 'Mo': 95.94, 'Tc': 98.9063, 'Ru': 101.07, 'Rh': 102.9055, 'Pd': 106.42,
    'Ag': 107.8682, 'Cd': 112.411, 'In': 114.818, 'Sn': 118.71, 'Sb': 121.760, 'Te': 127.6,
    'I': 126.90447, 'Xe': 131.293, 'Cs': 132.9054519, 'Ba': 137.327, 'La': 138.90547, 'Ce': 140.116,
    'Pr': 140.90465, 'Nd': 144.242, 'Pm': 146.9151, 'Sm': 150.36, 'Eu': 151.964, 'Gd': 157.25,
    'Tb': 158.92535, 'Dy': 162.5, 'Ho': 164.93032, 'Er': 167.259, 'Tm': 168.93421, 'Yb': 173.04,
    'Lu': 174.967, 'Hf': 178.49, 'Ta': 180.9479, 'W': 183.84, 'Re': 186.207, 'Os': 190.23, 'Ir': 192.217,
    'Pt': 195.084, 'Au': 196.966569, 'Hg': 200.59, 'Tl': 204.3833, 'Pb': 207.2, 'Bi': 208.9804,
    'Po': 208.9824, 'At': 209.9871, 'Rn': 222.0176, 'Fr': 223.0197, 'Ra': 226.0254, 'Ac': 227.0278,
    'Th': 232.03806, 'Pa': 231.03588, 'U': 238.02891, 'Np': 237.0482, 'Pu': 244.0642, 'Am': 243.0614,
    'Cm': 247.0703, 'Bk': 247.0703, 'Cf': 251.0796, 'Es': 252.0829, 'Fm': 257.0951, 'Md': 258.0951,
    'No': 259.1009, 'Lr': 262, 'Rf': 267, 'Db': 268, 'Sg': 271, 'Bh': 270, 'Hs': 269, 'Mt': 278,
    'Ds': 281, 'Rg': 281, 'Cn': 285, 'Nh': 284, 'Fl': 289, 'Mc': 289, 'Lv': 292, 'Ts': 294, 'Og': 294,
    '': 0
}


def get_molar_mass(compound: str, decimal_places=None) -> float:
    """
    Compute the molar mass based on a chemical formula.
    Adapted from https://gist.github.com/elibroftw/22e3b4c1eb7fa0a6c83d099d24200f95

    Parameters
    ----------
    compound : str
        Chemical formula to compute the molar mass of.
    decimal_places : int
        Number of decimal places to round the molar mass to.
    
    Returns
    -------
    out : float
        Molar mass of the compound.
    """
    is_polyatomic = end = multiply = False
    polyatomic_mass, m_m, multiplier = 0, 0, 1
    element = ''

    for e in compound:
        if is_polyatomic:
            if end:
                is_polyatomic = False
                m_m += int(e) * polyatomic_mass if e.isdigit() else polyatomic_mass + MM_of_Elements[e]
            elif e.isdigit():
                multiplier = int(str(multiplier) + e) if multiply else int(e)
                multiply = True
            elif e.islower():
                element += e
            elif e.isupper():
                polyatomic_mass += multiplier * MM_of_Elements[element]
                element, multiplier, multiply = e, 1, False
            elif e == ')':
                polyatomic_mass += multiplier * MM_of_Elements[element]
                element, multiplier = '', 1
                end, multiply = True, False
        elif e == '(':
            m_m += multiplier * MM_of_Elements[element]
            element, multiplier = '', 1
            is_polyatomic, multiply = True, False
        elif e.isdigit():
            multiplier = int(str(multiplier) + e) if multiply else int(e)
            multiply = True
        elif e.islower():
            element += e
        elif e.isupper():
            m_m += multiplier * MM_of_Elements[element]
            element, multiplier, multiply = e, 1, False
    m_m += multiplier * MM_of_Elements[element]
    if decimal_places is not None:
        return round(m_m, decimal_places)
    return m_m


import h5py
import torch
import numpy             as np
import matplotlib.pyplot as plt

from pomme.utils import interpolate
from tqdm           import tqdm
from time           import perf_counter


def make_object_with_len(obj):
    """
    Turn object into a tuple if it does not have a __len__ attribute.

    Parameters
    ----------
    obj: object
        Object to be turned into a tuple if it does not have a __len__ attribute.

    Returns
    -------
    obj: object
        Object turned into a tuple if it does not already had a __len__ attribute.
    """
    if hasattr(obj, "__len__"):
        return obj
    else:
        return (obj,)
        

class TensorModel():
    """
    A (deterministic) model in which every variable is represented by a PyTorch tensor.
    """
    reserved_keys = ["shape", "sizes", "free"]
    
    def __init__(self, sizes, shape, keys=[], dtau_warning_threshold=0.5):
        """
        Constructor for a Tensor model object.

        Parameters
        ----------
        sizes: array_like
            The sizes of the model box.
        shape: array_like
            The shape of the Tensor variables.
        keys: list (optional)
            List of keys for the model variables.
        dtau_warning_threshold: float (optional)
            Threshold for the optical depth at which a warning is issued.

        Raises
        ------
        ValueError
            If the sizes and shape have a different number of dimensions.
        """
        # Check the keys
        for key in keys:
            if key in TensorModel.reserved_keys:
                raise ValueError(f"{key} is a reserved key word, please choose a different one.")
        # Set sizes of the model box
        self.sizes = make_object_with_len(sizes)
        # Set shape of Tensor variables
        self.shape = tuple(make_object_with_len(shape))
        # Set the dimension of the model
        if len(self.shape) == len(self.sizes):
            self.dimension = len(self.shape)
        else:
            raise ValueError(f"sizes and shape have a different number of dimensions: {len(self.sizes)} and {len(self.shape)}")
        # Initialise variables with random Tensor with appropriate shape 
        self.vars = {v:torch.rand(self.shape, requires_grad=True, dtype=torch.float64) for v in keys}
        # Extract geometrical properties
        self.origin = None
        self.coords = None
        # Set dtau_warning_threshold
        self.dtau_warning_threshold = dtau_warning_threshold


    def deepcopy(self):
        """
        Create a deep copy of the TensorModel.
        """
        # Create a new TensorModel
        m = TensorModel(sizes=self.sizes, shape=self.shape, dtau_warning_threshold=self.dtau_warning_threshold)
        # Deepcopy all variables
        for key in self.keys():
            m[key] = self[key].detach().clone()
        # Return the new model
        return m


    def dim(self):
        """
        Getter for the dimension of the TensorModel.
        """
        return len(self.shape)
    

    def __getitem__(self, key):
        """
        Getter for variables (vars). Allows the use of [] operators.

        Parameters
        ----------
        key: str
            Key of the variable to be returned.

        Returns
        -------
        torch.Tensor
            The requested variable.
        """
        return self.vars[key]
    
    
    def __setitem__(self, key, value):
        """
        Setter for variables (vars). Allows the use of [] operators.

        Parameters
        ----------
        key: str
            Key of the variable to be set.
        value: torch.Tensor
            The value to be set.
        
        Raises
        ------
        ValueError
            If the key is a reserved key word.
        """
        if key in TensorModel.reserved_keys:
            raise ValueError(f"{key} is a reserved key word, please choose a different one.")
        if isinstance(value, torch.Tensor):
            self.vars[key] = value
        elif isinstance(value, np.ndarray):
            self.vars[key] = torch.from_numpy(value)
        else:
            self.vars[key] = torch.tensor(value)
        

    def parameters(self):
        """
        Return a list of all variables in the TensorModel.
        """
        return list(self.vars.values())
    
    
    def free_parameters(self):
        """
        Return a list of all "free" variables in the TensorModel.
        """
        return [v for v in self.vars.values() if v.requires_grad]
    
    
    def dx(self, i):
        """
        Return the size of a model element.

        Parameters
        ----------
        i: int
            Index of the dimension.

        Returns
        -------
        dx: float
            Size of the model element along axis or dimension i.
        """
        return self.sizes[i] / self.shape[i]
    
    
    def integrate(self, var, axis=0):
        """
        Integrate a variable along an axis of the model box.
        """
        return self.dx(axis) * torch.cumsum(var, dim=axis)
    
    
    def integrate_out(self, var, axis=0):
        """
        Integrate a variable out along an axis of the model box.
        """
        return self.dx(axis) * torch.sum(var, dim=axis)

    
    # def create_image(self, eta, chi, axis=0):
    #     """
    #     Formal solution of the transfer equation (discretised as TensorModel)
    #     """
    #     # Check for numercial issues.
    #     dtau_max = chi.max() * self.dx(axis)
    #     if dtau_max > self.dtau_warning_threshold:
    #         print('WARNING:')
    #         print(f'  dtau_max > {self.dtau_warning_threshold}, which might lead to numerical difficulties (dtau_max = {dtau_max})!')
    #     if chi.min() < 0.0:
    #         print('WARNING:')
    #         print(f'  Negative opacities encountered, which might lead to numerical difficulties!')
    #     # Compute the optical depth and the emerging intensity.
    #     tau = self.integrate    (chi,                 axis=axis)
    #     img = self.integrate_out(eta*torch.exp(-tau), axis=axis)
    #     return img
    

    def diff(self, arr, axis=-1):
        """
        Derivative along an axis.

        Parameters
        ----------
        arr: torch.Tensor
            The array for which to compute the derivative.
        axis: int (optional)
            The axis along which to compute the derivative.

        Returns
        -------
        diff: torch.Tensor
            The derivative along the specified axis.
        """
        # Prepend the one but first and append the one but last.
        # (Such that the first and last difference are the same.)
        ap0 = arr.index_select(dim=axis, index=torch.tensor([0]))
        ap1 = arr.index_select(dim=axis, index=torch.tensor([1]))
        am1 = arr.index_select(dim=axis, index=torch.tensor([arr.size(axis)-1]))
        am2 = arr.index_select(dim=axis, index=torch.tensor([arr.size(axis)-2]))
        pre = 2.0 * ap0 - ap1
        app = 2.0 * am1 - am2
        # Return the second-order difference.
        return 0.5 * (self.shape[axis] / self.sizes[axis]) * ( arr.diff(dim=axis, prepend=pre) + arr.diff(dim=axis, append=app) )


    def diff_x(self, arr):
        """
        Derivtive along x axis.

        Parameters
        ----------
        arr: torch.Tensor
            The array for which to compute the derivative.

        Returns
        -------
        diff_x: torch.Tensor
            The derivative along the x axis.
        """
        return self.diff(arr, axis=0)
    

    def diff_y(self, arr):
        """
        Derivtive along y axis.

        Parameters
        ----------
        arr: torch.Tensor
            The array for which to compute the derivative.

        Returns
        -------
        diff_y: torch.Tensor
            The derivative along the y axis.
        """
        return self.diff(arr, axis=1)

    
    def diff_z(self, arr):
        """
        Derivtive along z axis.

        Parameters
        ----------
        arr: torch.Tensor
            The array for which to compute the derivative.

        Returns
        -------
        diff_z: torch.Tensor
            The derivative along the z axis.
        """
        return self.diff(arr, axis=2)

    
    def keys(self):
        """
        Return the variable keys.
        """
        return self.vars.keys()


    def origin_as_index_array(self, origin):
        """
        Convert origin to a numpy array.

        Parameters
        ----------
        origin: array_like or str
            Indices of the origin of the coordinate system (can be float).
            The dimension of the origin should match the dimension of the model.
            If 'centre', the origin is set to the centre of the model.

        Returns
        -------
        origin: numpy array
            indices of the origin of the coordinate system.
        """
        if isinstance(origin, str) and (origin == 'centre'):
            return 0.5 * (np.array(self.shape) - 1)
        else:
            return np.array(origin)
    
    
    def get_coords(self, origin='centre'):
        """
        Getter for the coordinates of each tensor location.
        
        Parameters
        ----------
        origin: array_like
            Indices of the origin of the coordinate system (can be float).
            The dimension of the origin should match the dimension of the model.
            If 'centre', the origin is set to the centre of the model.

        Returns
        -------
        coords: numpy array
            coordinates of each location in the tensor model.
        """
        # Cast origin into a numpy array
        self.origin = self.origin_as_index_array(origin)
        # Compute the coordinates of each tensor location
        if self.dimension == 1:
            self.coords = np.arange(self.shape[0])
        else:
            self.coords = np.moveaxis(np.indices(self.shape), 0, -1)
        # Shift coordinates w.r.t. origin
        self.coords = self.coords - self.origin
        # Scale coordinates to box size
        self.coords = np.moveaxis(self.coords * (np.array(self.sizes)/np.array(self.shape)), -1, 0)
        return self.coords

    
    def get_radius(self, origin='centre'):
        """
        Getter for the radial cooridnate of each location.

        Parameters
        ----------
        origin: array_like
            indices of the origin of the coordinate system (can be float).
            The dimension of the origin should match the dimension of the model.
            If 'centre', the origin is set to the centre of the model.
        
        Returns
        -------
        radius: numpy array
            radial coordinate of each location in the tensor model. 
        """
        # Check if coords are already set
        if (self.coords is None) or not np.all(self.origin_as_index_array(origin) == self.origin):
            self.coords = self.get_coords(origin)
        # In 1D there only is a radial coord
        if self.dimension == 1:
            return np.abs(self.coords)
        else:
            return np.linalg.norm(self.coords, axis=0)
        

    def get_radial_direction(self, origin='centre'):
        """
        Getter for the radial direction of the model.

        Parameters
        ----------
        origin: array_like (optional)
            indices of the origin of the coordinate system (can be float).
            The dimension of the origin should match the dimension of the model.

        Returns
        -------
        direction: numpy array
            radial directions of each element in the model.
        """    
        coords = self.get_coords(origin)
        radius = self.get_radius(origin)

        direction = (1.0 / radius) * coords
        
        return direction
        
        
    def apply(self, func, exclude=[], include=[]):
        """
        Apply the given functional to all model variables.

        Parameters
        ----------
        func: function
            The function to be applied to the model variables.
        exclude: list (optional)
            List of keys to exclude from the application of the function.
        include: list (optional)
            List of variables to include in the application of the function.
        
        Returns
        -------
        res: torch.Tensor
            The result of the application of the function to the model variables.
        """
        res = torch.zeros(1)
        for key in self.keys():
            if key not in exclude:
                res += func(self[key])
        for var in include:
            res += func(var)
        return res

    
    def apply_to_fields(self, func, exclude=[], include=[]):
        """
        Apply the given functional to all model fields.

        Parameters
        ----------
        func: function
            The function to be applied to the model fields.
        exclude: list (optional)
            List of keys to exclude from the application of the function.
        include: list (optional)
            List of variables to include in the application of the function.

        Returns
        -------
        res: torch.Tensor
            The result of the application of the function to the model fields.
        """
        res = torch.zeros(1)
        for key in self.keys():
            if self.is_field(key) and (key not in exclude):
                res += func(self[key])
        for var in include:
            res += func(var)
        return res
    
    
    def free(self, keys):
        """
        "Frees" the variables that should be adjusted in optimistation.
        (and hence require a gradient.)

        Parameters
        ----------
        keys: str or list
            The key(s) of the variable(s) that should be freed.
        """
        # Make sure that keys are iterable
        if not isinstance(keys, list):
            keys = [keys]
        # Free variables 
        for key in keys:
            self.vars[key].requires_grad_(True)
            
            
    def free_all(self):
        """
        Makes all model variables free.
        """
        self.free(list(self.keys()))


    def fix(self, keys):
        """
        "Fixes" the variables that should not be adjusted in optimistation.
        (and hence do not require a gradient.)

        Parameters
        ----------
        keys: str or list
            The key(s) of the variable(s) that should be fixed.
        """
        # Make sure that keys are a list
        if not isinstance(keys, list):
            keys = [keys]
        # Free variables 
        for key in keys:
            self.vars[key].requires_grad_(False)
    
    def fix_all(self):
        """
        Makes all model variables fixed.
        """
        self.fix(list(self.keys()))
            
            
    def is_field(self, key):
        """
        Check if the key corresponds to a field (i.e. a variable defined at each model element).

        Parameters
        ----------
        key: str
            The key of the variable.

        Returns
        -------
        is_field: bool
            True if the key corresponds to a field, False otherwise.
        """
        return self[key].shape == self.shape
        
        
    def get_detached_clone(self):
        """
        Returns a copy of the model, detached from this model. 
        """
        # Create a new TensorModel
        model = TensorModel(sizes=self.sizes, shape=self.shape)
        # Add detached clones of all variables
        for key in self.keys():
            model[key] = self[key].detach().clone()
        # Return the new model
        return model
    
    
    def info(self):
        """
        Print info about the TensorModel.
        """
        print("Variable key:              Free/Fixed:   Field:    Min:           Mean:          Max:")
        for key in self.keys():
            var = self[key]
            if var.requires_grad:
                fof = "Free"
            else:
                fof = "Fixed"
            if self.is_field(key):
                field = "True"
            else:
                field = "False"
            print(f"  {key:<25}  {fof:<5}         {field:<5}    {var.min():+.3e}     {var.mean():+.3e}     {var.max():+.3e}")
        print("sizes:", self.sizes)
        print("shape:", self.shape)

        
    @staticmethod
    def print_diff(model_1, model_2):
        """
        Print info about the difference between two models.

        Parameters
        ----------
        model_1: TensorModel
            The first model.
        model_2: TensorModel
            The second model.
        """
        print("Variable:                 Min:          Mean:         Max:")
        for key in model_1.keys():
            diff = torch.abs(model_1[key] - model_2[key])
        print(f"  {key:<25} {diff.min():.3e}     {diff.mean():.3e}     {diff.max():.3e}")


    def save(self, fname):
        """
        Save the TensorModel to disk as an HDF5 file.

        Parameters
        ----------
        fname: str
            The name of the file to which the TensorModel is saved.
        """
        def overwrite_dataset(file, key, data):
            try:
                del file[key]
            except:
                pass
            file[key] = data
        # Open or create an HDF5 file
        with h5py.File(fname, "a") as file:
            # Store the shape and sizes
            overwrite_dataset(file, "shape", self.shape)
            overwrite_dataset(file, "sizes", self.sizes)
            # Create a list to store the keys of the free variables
            free = []
            # Store the variables
            for key in self.keys():
                overwrite_dataset(file, key, self[key].data)
                # Check if it is a free variables
                if self[key].requires_grad:
                    # If so, store its key
                    free.append(key)
            # Store the keys of the free variables
            overwrite_dataset(file, "free", free)
        
        
    @staticmethod
    def load(fname):
        """
        Load a TensorModel from an HDF5 file.

        Parameters
        ----------
        fname: str
            The name of the file from which the TensorModel is loaded.

        Returns
        -------
        model: TensorModel
            The loaded TensorModel
        """
        with h5py.File(fname, "r") as file:
            # Read the shape and sizes
            shape = np.array(file["shape"])
            sizes = np.array(file["sizes"])
            # Create a TensorModel object
            model = TensorModel(shape=shape, sizes=sizes)
            # Read the list of keys of the free variables
            free = [f.decode("utf-8") for f in np.array(file["free"])]
            # Read the variables
            for key in file.keys():
                if key not in TensorModel.reserved_keys:
                    model[key] = np.array(file[key])
                    if key in free:
                        model[key].requires_grad_(True)
        # Return loaded model
        return model

    
    def interpolate_from_model(self, model, detach=True):
        """
        Interpolate the varialbles of another model into this model.

        Parameters
        ----------
        model: TensorModel
            The model from which to interpolate the variables.
        detach: bool (optional)
            If True, the interpolated tensors in this model will be detached
            from the PyTorch compute graph.
        """
        for key in model.keys():
            if model.is_field(key):
                self[key] = interpolate(model[key], self.shape).clone()
            else:
                self[key] = model[key].clone()
            # If required, detach tensor from the compute graph
            if detach:
                self.vars[key].detach_()
        


# class SphericallySymmetric():
#     """
#     Spherically Symmetric model functionality.
#     """
#     def __init__(self, model_1D):
#         """
#         Constructor for a spherically symmetric model from a 1D TensorModel.
#         """
#         # Check if the input model is in fact 1D
#         if model_1D.dimension != 1:
#             raise ValueError(f"Input model is not 1D. model_1D.dimension = {model_1D.dimension}.")

#         # Define the underlying 1D and 2D models
#         self.model_1D = model_1D
#         self.model_2D = TensorModel(
#             sizes=(2*self.model_1D.sizes[0], self.model_1D.sizes[0]),
#             shape=(2*self.model_1D.shape[0], self.model_1D.shape[0])
#         )
        
#         # Set origin for the 2D model halfway along the x axis
#         self.origin_2D = np.array([0.5*(self.model_2D.shape[0]-1), 0.0])

#         # Get linear interpolation utilities
#         self.i_min, self.i_max, self.l = self.linint_utils(self.model_2D, self.origin_2D)
    
#         # Get the circular weights
#         self.c_weights = self.get_circular_weights()


#     def linint_utils(self, model, origin):
#         """
#         Getter for the linear interpolation utilities.
#         """
#         # Compute the radii in the 2D model
#         rs = torch.from_numpy(model.get_radius(origin=origin))

#         # Compute the 2D radii in the 1D model
#         rs = (self.model_1D.shape[0] / self.model_1D.sizes[0]) * rs

#         # Compute the corresponding indices of the 2D radii in the 1D model
#         i_min = torch.floor(rs).type(torch.int64)
#         i_max = i_min + 1

#         # Bound the indices from below
#         i_min[i_min < 0] = 0
#         i_max[i_max < 0] = 0

#         # Bound the indices form above
#         i_min[i_min >= self.model_1D.shape[0]] = self.model_1D.shape[0]-1
#         i_max[i_max >= self.model_1D.shape[0]] = self.model_1D.shape[0]-1

#         # Compute the scaling factor
#         l = rs - i_min.type(torch.float64)
        
#         return i_min, i_max, l
        
         
#     def map_variables(self, keys):
#         """
#         Map a 1D Tensor Model variable to a 2D TensorModel variable assuming spherical symmetry.
#         """    
#         # Precompute the linear map
#         a = self.l
#         b = 1.0 - self.l
#         # Map variables
#         for key in keys:
#             if self.model_1D.is_field(key):
#                 self.model_2D[key] = a*self.model_1D[key][self.i_max] + b*self.model_1D[key][self.i_min]
#             else:
#                 self.model_2D[key] = self.model_1D[key]
 

#     def map_1D_to_2D(self):
#         """
#         Map a 1D TensorModel to a 2D TensorModel assuming spherical symmetry.
#         """    
#         self.map_variables(self.model_1D.keys())

        
#     def create_3D_model(self, sizes, shape, origin, detach=True):
#         """
#         Create a 3D TensorModel for the current spherically symmetric model, and given an origin.
#         If detach=True, as is by default, all model tensors will be detached form the the spherical model.
#         """
#         # Check sizes
#         if len(sizes) != 3:
#             raise ValueError("Provided sizes do not correspond to a 3D model.")
#         # Check shape
#         if len(shape) != 3:
#             raise ValueError("Provided shape does not correspond to a 3D model.")
#         # Create a 3D tensor model
#         model_3D = TensorModel(sizes=sizes, shape=shape)
#         # Compute the linear integration utilities
#         i_min, i_max, l = self.linint_utils(model_3D, origin)
#         # Precompute the linear map
#         a = l
#         b = 1.0 - l
#         # Map variables
#         for key in self.model_1D.keys():
#             if self.model_1D.is_field(key):
#                 model_3D[key] = a*self.model_1D[key][i_max] + b*self.model_1D[key][i_min]
#             else:
#                 model_3D[key] = self.model_1D[key].clone()
#             # If required, detach tensor from the compute graph
#             if detach:
#                 model_3D[key].detach_()
#         # Return the 3D TensorModel
#         return model_3D

    
#     def get_circular_weights(self):
#         """
#         Getter for the circular weights.
#         """
#         coords = self.model_2D.get_coords(origin=self.origin_2D)
#         r = coords[1,0]
#         r[0] = 0.5 * r[1]
#         return torch.from_numpy(2.0*np.pi*self.model_1D.dx(0)*r)
    
    
#     def integrate_intensity(self, img):
#         """
#         Computes the integrated intensity.
#         """
#         return torch.einsum("if,i -> f", img, self.c_weights)
    

#     def plot(self, keys=None, exclude=[]):
#         """
#         Plot the model parameters.
#         """
#         # Pick which variables to plot
#         if keys is None:
#             keys = self.model_1D.keys()
#         # Get the radial coordinate of the model
#         r = self.model_1D.get_radius(origin=[0])
#         # Plot each 1D model variable
#         for key in keys:
#             if key not in exclude:
#                 if self.model_1D.is_field(key):
#                     plt.figure(dpi=130)
#                     plt.plot(r, self.model_1D[key].data)
#                     plt.xlabel('r')
#                     plt.ylabel(key)
#                 else:
#                     print(f"{key:<21}{self.model_1D[key].data}")
        

class SphericalModel:
    """
    Spherically symmetric model.
    Convenience class to simplify creating spherically symmetic models.
    """
    def __init__(self, rs, model_1D, r_star=0.0):
        """
        Constructor for a spherically symmetric model.

        Parameters
        ----------
        rs: array_like
            The radii of the spherical model.
        model_1D: TensorModel
            The 1D model for which to construct the spherical model.
        r_star: float (optional)
            The radius of the star. (Inner boundary of the spherical model.)
        """
        self.rs       = rs            # Radii of the spherical model
        self.Nb       = len(rs) - 1   # Number of impact parameters
        self.r_star   = r_star        # Radius of the star
        self.model_1D = model_1D

        # Setup rays
        self.image_ray_tracer()


    def diff_r(self, arr, rs):
        """
        Derivative along the radial direction.

        Parameters
        ----------
        arr: torch.Tensor
            The array for which to compute the derivative.
        rs: torch.Tensor
            The radii of the spherical model.

        Returns
        -------
        diff_r: torch.Tensor
            The derivative along the radial direction.
        """
        # Prepend the one but first and append the one but last.
        # (Such that the first and last difference are the same.)
        ap0 = arr.index_select(dim=0, index=torch.tensor([0]))
        ap1 = arr.index_select(dim=0, index=torch.tensor([1]))
        am1 = arr.index_select(dim=0, index=torch.tensor([arr.size(0)-1]))
        am2 = arr.index_select(dim=0, index=torch.tensor([arr.size(0)-2]))

        pre = 2.0 * ap0 - ap1
        app = 2.0 * am1 - am2

        dr = rs.diff(prepend=torch.zeros(1))

        # Return the second-order difference.
        return (0.5 / dr) * ( arr.diff(prepend=pre) + arr.diff(append=app) )


    def get_velocity(self, model_1D):
        """
        Getter for the velocity field.
        Has to be implemented by the user!
        """
        raise NotImplementedError('First implement and define get_velocity.')


    def get_temperature(self, model_1D):
        """
        Getter for the temperature distribution.
        Has to be implemented by the user!
        """
        raise NotImplementedError('First implement and define get_temperature.')


    def get_abundance(self, model_1D):
        """
        Getter for the abundance distribution.
        Has to be implemented by the user!
        """
        raise NotImplementedError('First implement and define get_abundance.')


    def get_turbulence(self, model_1D):
        """
        Getter for the turbulence distribution.
        Has to be implemented by the user!
        """
        raise NotImplementedError('First implement and define get_turbulence.')


    def get_boundary_condition(self, model_1D, frequency, b):
        """
        Getter for the boundary condition at the impact parameter b.
        Has to be implemented by the user!
        """
        raise NotImplementedError('First implement and define get_boundary_condition.')


    def image_ray_tracer(self):
        """
        Ray tracer for the spherical model.
        """
        # Initialise lists that will contain results
        self.dZss = []   # Distance increments along th ray
        self.idss = []   # Grid point indices along the ray
        self.diss = []   # Directional cosinses along the ray

        # For each impact parameter
        for i in range(self.Nb):

            Zs  = np.sqrt((self.rs[i:] - self.rs[i]) * (self.rs[i:] + self.rs[i]))
            dZs = np.diff(Zs)
            ids = np.arange(i+1, self.Nb+1)
            dis = Zs[1:] / self.rs[i+1:]

            # Check for NaNs (may occur by cancellation errors in the definition of Zs)
            if np.isnan(dZs).any():
                raise Warning('NaNs in dZs!')
            if np.isnan(dis).any():
                raise Warning('NaNs in dis!')

            if self.rs[i] >= self.r_star:
                # Append the data to the lists for this impact parameter
                self.dZss.append(torch.from_numpy(np.concatenate(( dZs[::-1],        dZs))))
                self.idss.append(torch.from_numpy(np.concatenate(( ids[::-1], [i],   ids))))
                self.diss.append(torch.from_numpy(np.concatenate((-dis[::-1], [0.0], dis))))
            else:
                # Find where the (reverse) ray is outside the star
                # Note: we need the reverse ray since our observer is at -R
                mask = (self.rs[i+1:][::-1] > self.r_star)
                # Append the data to the lists for this impact parameter
                self.dZss.append(torch.from_numpy(dZs[::-1][mask][:-1]))
                self.idss.append(torch.from_numpy(ids[::-1][mask]))
                self.diss.append(torch.from_numpy(dis[::-1][mask]))


    def image(self, lines, frequencies, r_max=np.inf, step=5):
        """
        Create synthetic image of the given spherically symmetric model.

        Parameters
        ----------
        lines: list
            List of Line objects for which to create the synthetic image.
        frequencies: list
            List of frequencies at which to compute the synthetic image.
        r_max: float (optional)
            Maximum radius at which to compute the synthetic image.
        step: int (optional)
            Step size for the impact parameters at which to trace a ray.

        Returns
        -------
        Iss: torch.Tensor
            The synthetic image of the model.
        """
        # Extract the model parameters
        velocity    = self.get_velocity   (self.model_1D)
        abundance   = self.get_abundance  (self.model_1D)
        temperature = self.get_temperature(self.model_1D)
        turbulence  = self.get_turbulence (self.model_1D)

        # Tensor for the intensities in each line
        Iss = torch.zeros((len(lines), len(frequencies[0])), dtype=torch.float64)

        # For each line
        for l, (line, freq) in enumerate(zip(lines, frequencies)):

            # Check that the number of frequencies is the same for all lines
            assert len(freq) == len(Iss[l])

            b_prev = 0.0

            # For each impact parameter
            for i in range(0, self.Nb, step):
                
                if self.rs[i] < r_max:

                    # Impact parameter
                    b = self.rs[i]

                    # Get boundary condition at this impact parameter
                    img_bdy = self.get_boundary_condition(self.model_1D, frequency=freq, b=b)

                    # Get intensity at this impact parameter
                    I_loc = line.LTE_image_along_last_axis(
                        abundance    = abundance  [self.idss[i]],
                        temperature  = temperature[self.idss[i]],
                        v_turbulence = turbulence [self.idss[i]],
                        velocity_los = velocity   [self.idss[i]] * self.diss[i],
                        frequencies  = freq,
                        dx           = self.dZss[i],
                        img_bdy      = img_bdy
                    )

                    # Surface area of the annulus at each impact parameter
                    dss = np.pi * (b + b_prev) * (b - b_prev)

                    # Integrate this piece of the annulus
                    Iss[l] += dss * I_loc

                    b_prev = b
                    
        return Iss


    def plot(self, keys=None, exclude=[]):
        """
        Plot the model parameters.
        """
        # Pick which variables to plot
        if keys is None:
            keys = self.model_1D.keys()
        # Get the radial coordinate of the model
        r = self.model_1D.get_radius(origin=[0])
        # Plot each 1D model variable
        for key in keys:
            if key not in exclude:
                if self.model_1D.is_field(key):
                    plt.figure(dpi=130)
                    plt.plot(r, self.model_1D[key].data)
                    plt.xlabel('r')
                    plt.ylabel(key)
                else:
                    print(f"{key:<21}{self.model_1D[key].data}")
        


class GeneralModel:
    """
    General model class.
    """
    def __init__(self, model):
        """
        Constructor for a general model.

        Parameters
        ----------
        model: TensorModel
            The model for which to create the general model.
        """
        self.model = model


    def get_velocity(self, model):
        """
        Getter for the velocity field.
        Has to be implemented by the user!
        """
        raise NotImplementedError('First implement and define get_velocity.')


    def get_temperature(self, model):
        """
        Getter for the temperature distribution.
        Has to be implemented by the user!
        """
        raise NotImplementedError('First implement and define get_temperature.')


    def get_abundance(self, model, l):
        """
        Getter for the abundance distribution (for the line produce species of line l).
        Has to be implemented by the user!
        """
        raise NotImplementedError('First implement and define get_abundance.')


    def get_turbulence(self, model):
        """
        Getter for the turbulence distribution.
        Has to be implemented by the user!
        """
        raise NotImplementedError('First implement and define get_turbulence.')


    def get_boundary_condition(self, model, frequency):
        """
        Getter for the boundary condition.
        Has to be implemented by the user!
        """
        raise NotImplementedError('First implement and define get_boundary_condition.')


    # Forward model
    def image(self, lines, frequencies):
        """
        Create synthetic image of the given general model.

        Parameters
        ----------
        lines: list
            List of Line objects for which to create the synthetic image.
        frequencies: list

        Returns
        -------
        Iss: torch.Tensor
            The synthetic image of the model.
        """
        # Tensor for the intensities in each line
        imgs = torch.zeros((len(lines), self.model.shape[0], self.model.shape[1], len(frequencies[0])), dtype=torch.float64)

        # For each line
        for l, (line, freq) in enumerate(zip(lines, frequencies)):

            # Check that the number of frequencies is the same for all lines
            assert len(freq) == imgs[l].shape[-1]

            imgs[l] = line.LTE_image_along_last_axis(
                abundance    = self.get_abundance  (self.model, l),
                temperature  = self.get_temperature(self.model),
                v_turbulence = self.get_turbulence (self.model),
                velocity_los = self.get_velocity   (self.model)[2],
                frequencies  = freq,
                dx           = self.model.dx(0),
                img_bdy      = self.get_boundary_condition(self.model, freq)
            )

        return imgs


import numpy as np
import numba


class Haar():

    def __init__(self, points, q):
        """
        Constructor for a Haar object.

        Parameters
        ----------
        points    : n by 3 array 
        q         : Maximum level of refinement (side of the cube is 2**q, number of volume elements 8**q)
        """
        # Store the maximum (finenst) level
        self.q = q
        
        # Compute size of the box
        self.xyz_min = np.min(points, axis=0)
        self.xyz_max = np.max(points, axis=0)

        # Set a tolerance (which should be smaller than the resolution of the cube)
        tol = 1.0e-3 / 2**self.q

        # Compute xyz size of the box (1 tol'th larger)
        self.xyz_L = (1.0 + tol) * (self.xyz_max - self.xyz_min)

        # Normalize the point coordinates to be in [0,1]
        self.points_normed = (points - self.xyz_min) / self.xyz_L

        # Count the number of points in each octree grid cell
        self.num = Haar.get_number_matrices(self.points_normed, self.q)

        # Compute cube (ix, iy, iz) indices in the linear representation 
        self.ids = [np.array(Haar.__cub__(np.arange(8**k, dtype=np.int64)), dtype=np.int64).T for k in range(q)]


    def map_data(self, data, interpolate=True):
        """
        Map data to the cube.
        """
        # Integrate (i.e. sum) the point data over each cell
        dat = Haar.integrate_data(self.points_normed, self.q, data)
        # Divide by the numberof points in each cell to get the average
        for k in range(self.q):
            dat[k] = np.divide(dat[k], self.num[k], out=dat[k], where=(self.num[k]!=0))
        # Interpolate empty cells in each cube
        if interpolate:
            for k in range(1, self.q):
                dat[k] = Haar.interpolate(self.ids[k], self.num[k], dat[k], dat[k-1], k)
        # Return a hierarchical representation of the data
        return dat


    @staticmethod
    @numba.njit(parallel=True)
    def get_number_matrices(points_normed, q):
        """
        Compute the number of points that live in each cell at each level.
        """
        # Create number matrices
        num = [np.zeros((2**k, 2**k, 2**k), dtype=np.int64) for k in range(q)]
        # For all levels
        for k in range(q):
            # Compute the indices of the points in the octree grid
            indices = (points_normed * 2**k).astype(np.int64)
            # Count the number of points at every index
            for ix, iy, iz in indices:
                num[k][ix, iy, iz] += 1
        # Return the number matrices
        return num

    
    @staticmethod
    @numba.njit(parallel=True)
    def integrate_data(points_normed, q, data):
        """
        Sum the data in each cell at each level.
        """
        # Create number matrices
        dat = [np.zeros((2**k, 2**k, 2**k), dtype=data.dtype) for k in range(q)]
        # For all levels
        for k in range(q):
            # Compute the indices of the points in the octree grid
            indices = (points_normed * 2**k).astype(np.int64)
            # Count the number of points at every index
            for i, (ix, iy, iz) in enumerate(indices):
                dat[k][ix, iy, iz] += data[i]
        # Return the integrated data
        return dat


    @staticmethod
    @numba.njit(parallel=True)
    def interpolate(ids, num, dat, dat_up, k):
        # For all cells in the cube at this level
        for i in numba.prange(8**k):
            # Get the cube triple index
            ix, iy, iz = ids[i]
            # If the cell at this level is empty take the value from the above level
            if (num[ix, iy, iz] == 0):
                dat[ix, iy, iz] = dat_up[ix//2, iy//2, iz//2]
        # Return data cube
        return dat
    
    
    @staticmethod
    @numba.njit(parallel=True)
    def __lin__(i, j, k):
        """
        If the binary representation of i, j, and k are given by:
            bin(i) = (..., i8, i4, i2, i1)
            bin(j) = (..., j8, j4, j2, j1)
            bin(k) = (..., k8, k4, k2, k1)
        this function returns the number, r, for which the binary repressentation is given by:
            bin(r) = (..., k8, j8, i8, k4, j4, i4, k2, j2, i2, k1, j1, i1)
        inducing an hierarchical ordering.
        """

        r = 0

        j = j << 1
        k = k << 2

        r = r + (i & 2**0)
        r = r + (j & 2**1)
        r = r + (k & 2**2)

        # Yes, this can be rewritten as a for-loop
        # Done this way since I am not sure if the compiler unrolls loops.
        # Should be tested!

        i = i << 2
        j = j << 2
        k = k << 2

        r = r + (i & 2**3)
        r = r + (j & 2**4)
        r = r + (k & 2**5)

        i = i << 2
        j = j << 2
        k = k << 2

        r = r + (i & 2**6)
        r = r + (j & 2**7)
        r = r + (k & 2**8)

        i = i << 2
        j = j << 2
        k = k << 2

        r = r + (i & 2**9)
        r = r + (j & 2**10)
        r = r + (k & 2**11)

        i = i << 2
        j = j << 2
        k = k << 2

        r = r + (i & 2**12)
        r = r + (j & 2**13)
        r = r + (k & 2**14)

        i = i << 2
        j = j << 2
        k = k << 2

        r = r + (i & 2**15)
        r = r + (j & 2**16)
        r = r + (k & 2**17)

        i = i << 2
        j = j << 2
        k = k << 2

        r = r + (i & 2**18)
        r = r + (j & 2**19)
        r = r + (k & 2**20)

        i = i << 2
        j = j << 2
        k = k << 2

        r = r + (i & 2**21)
        r = r + (j & 2**22)
        r = r + (k & 2**23)

        i = i << 2
        j = j << 2
        k = k << 2

        r = r + (i & 2**24)
        r = r + (j & 2**25)
        r = r + (k & 2**26)

        i = i << 2
        j = j << 2
        k = k << 2

        r = r + (i & 2**27)
        r = r + (j & 2**28)
        r = r + (k & 2**29)

        i = i << 2
        j = j << 2
        k = k << 2

        r = r + (i & 2**30)
        r = r + (j & 2**31)
        r = r + (k & 2**32)

        return r


    @staticmethod
    @numba.njit(parallel=True)
    def __cub__(r):
        """
        If the binary representation of r is given by:
            bin(r) = (..., k8, j8, i8, k4, j4, i4, k2, j2, i2, k1, j1, i1)
        this function returns the numbers, i, j, and k, for which the binary repressentations are given by:
            bin(i) = (..., i8, i4, i2, i1)
            bin(j) = (..., j8, j4, j2, j1)
            bin(k) = (..., k8, k4, k2, k1)
        inducing an hierarchical ordering.
        """

        i = 0
        j = 0
        k = 0

        # Yes, this can be rewritten as a for-loop
        # Done this way since I am not sure if the compiler unrolls loops.
        # Should be tested!

        i = i + (r & 2**0)
        r = r >> 1
        j = j + (r & 2**0)
        r = r >> 1
        k = k + (r & 2**0)

        i = i + (r & 2**1)
        r = r >> 1
        j = j + (r & 2**1)
        r = r >> 1
        k = k + (r & 2**1)

        i = i + (r & 2**2)
        r = r >> 1
        j = j + (r & 2**2)
        r = r >> 1
        k = k + (r & 2**2)

        i = i + (r & 2**3)
        r = r >> 1
        j = j + (r & 2**3)
        r = r >> 1
        k = k + (r & 2**3)

        i = i + (r & 2**4)
        r = r >> 1
        j = j + (r & 2**4)
        r = r >> 1
        k = k + (r & 2**4)

        i = i + (r & 2**5)
        r = r >> 1
        j = j + (r & 2**5)
        r = r >> 1
        k = k + (r & 2**5)

        i = i + (r & 2**6)
        r = r >> 1
        j = j + (r & 2**6)
        r = r >> 1
        k = k + (r & 2**6)

        i = i + (r & 2**7)
        r = r >> 1
        j = j + (r & 2**7)
        r = r >> 1
        k = k + (r & 2**7)

        i = i + (r & 2**8)
        r = r >> 1
        j = j + (r & 2**8)
        r = r >> 1
        k = k + (r & 2**8)

        i = i + (r & 2**9)
        r = r >> 1
        j = j + (r & 2**9)
        r = r >> 1
        k = k + (r & 2**9)

        i = i + (r & 2**10)
        r = r >> 1
        j = j + (r & 2**10)
        r = r >> 1
        k = k + (r & 2**10)

        i = i + (r & 2**11)
        r = r >> 1
        j = j + (r & 2**11)
        r = r >> 1
        k = k + (r & 2**11)

        i = i + (r & 2**12)
        r = r >> 1
        j = j + (r & 2**12)
        r = r >> 1
        k = k + (r & 2**12)

        i = i + (r & 2**13)
        r = r >> 1
        j = j + (r & 2**13)
        r = r >> 1
        k = k + (r & 2**13)

        i = i + (r & 2**14)
        r = r >> 1
        j = j + (r & 2**14)
        r = r >> 1
        k = k + (r & 2**14)

        i = i + (r & 2**15)
        r = r >> 1
        j = j + (r & 2**15)
        r = r >> 1
        k = k + (r & 2**15)

        return i, j, k